/********************************************************************************************
 *                                                                                          *
 * Copyright (C) 2017 Armin Felder, Dennis Beier                                            *
 * This file is part of RocketChatMobileEngine <https://git.fairkom.net/chat/fairchat>.     *
 *                                                                                          *
 * RocketChatMobileEngine is free software: you can redistribute it and/or modify           *
 * it under the terms of the GNU General Public License as published by                     *
 * the Free Software Foundation, either version 3 of the License, or                        *
 * (at your option) any later version.                                                      *
 *                                                                                          *
 * RocketChatMobileEngine is distributed in the hope that it will be useful,                *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                           *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                            *
 * GNU General Public License for more details.                                             *
 *                                                                                          *
 * You should have received a copy of the GNU General Public License                        *
 * along with RocketChatMobileEngine. If not, see <http://www.gnu.org/licenses/>.           *
 *                                                                                          *
 ********************************************************************************************/


#include "rocketchatserver.h"



RocketChatServerData::RocketChatServerData( const QString &pId,  QString pBaseUrl,
        QString pApiUri,
        UserModel &pUserModel,
        ChannelModel &channelModel,
        ChannelModel &groupsModel,
        ChannelModel &directModel,
        MessagesModel &messagesModel,
        LoginMethodsModel &loginMethodsModel )
    : mBaseUrl( pBaseUrl ), mApiUri( pApiUri ), mServerId( pId ), userModel( pUserModel ),
      channelsModel( channelModel ), groupsModel( groupsModel ), directModel( directModel ), mMessagesModel( messagesModel ), loginMethodsModel( loginMethodsModel )
{
    mStorage = PersistanceLayer::instance();
    mEmojiRepo = new EmojiRepo;
    historyLoaded = [ = ]( QMultiMap<QString, QSharedPointer<RocketChatMessage>> *messages ) {
        if ( messages ) {
            mStorage->transaction();

            for ( const auto &key : messages->uniqueKeys() ) {
                if ( Q_LIKELY( mChannels->contains( key ) && !mChannels->get( key ).isNull() ) ) {
                    MessageList messageList = messages->values( key );
                    mChannels->get( key )->addMessages( messageList );
                    mMessageService->persistMessages( messageList );
                }
            }

            mStorage->commit();

            if ( messages != nullptr ) {
                delete messages;
            }
        }
    };
    mFileService = new FileService( this );
    mEmojiService = new EmojiService( this, mFileService, mStorage );
    setRestApi( new RestApi( this, QStringLiteral( "https://" ) + mBaseUrl, mApiUri ) );
}

void RocketChatServerData::init()
{
    mDdpApi = new MeteorDDP( this, mBaseUrl );

    mRestApi->moveToThread( QThread::currentThread() );
    mDdpApi->moveToThread( QThread::currentThread() );
    mUsername = mStorage->getUserName();

    mChannels = new ChannelRepository( channelsModel, channelsModel, groupsModel, mMessagesModel );

    QMetaObject::invokeMethod(mRestApi,"init");

    connect( mDdpApi, &MeteorDDP::messageReceived, this, &RocketChatServerData::onDDPMessageReceived, Qt::UniqueConnection );
    connect( mDdpApi, &MeteorDDP::ddpConnected, this, &RocketChatServerData::onDDPConnected, Qt::UniqueConnection );
    connect( mDdpApi, &MeteorDDP::ddpDisconnected, this, &RocketChatServerData::onDDPDisonnected, Qt::UniqueConnection );

    mUserId = mStorage->getUserId();
    mUsername = mStorage->getUserName();

    // mEmojisMap.reserve(3000);
    mDownloadsInProgress.reserve( 100 );

    loadEmojis();
    //mMessageService = new MessageService( mStorage, this, mEmojisMap );
    mMessageService = new MessageService( mStorage, this, mEmojiRepo, &mMessagesModel );

    mChannelService = new RocketChatChannelService( this, mMessageService );
    mChannelService->setDdp( mDdpApi );
    mChannelService->setChannels( mChannels );

    connect( mChannelService, &RocketChatChannelService::channelsLoaded, this, &RocketChatServerData::onChannelsLoaded, Qt::UniqueConnection );
    connect( mChannelService, &RocketChatChannelService::usersLoaded, this, &RocketChatServerData::onUsersLoaded, Qt::UniqueConnection );
    connect( mChannelService, &RocketChatChannelService::directChannelReady, this, &RocketChatServerData::switchChannelByRid, Qt::UniqueConnection );
    connect( &channelsModel, &ChannelModel::unreadMessagesChanged, this, &RocketChatServerData::onUnreadCountChanged, Qt::UniqueConnection );
    connect( &directModel, &ChannelModel::unreadMessagesChanged, this, &RocketChatServerData::onUnreadCountChanged, Qt::UniqueConnection );
    connect( &groupsModel, &ChannelModel::unreadMessagesChanged, this, &RocketChatServerData::onUnreadCountChanged, Qt::UniqueConnection );


    QString lastServer = mStorage->getSetting( QStringLiteral( "currentServer" ) );

    if ( lastServer == mBaseUrl ) {
        QPair<QString, uint> tokenDb = mStorage->getToken();
        QString userDb = mStorage->getUserName();
        QDateTime currentTime = QDateTime::currentDateTime();

        if ( !tokenDb.first.isEmpty() && tokenDb.second > currentTime.toTime_t() ) {
            loginWithToken( userDb, tokenDb.first );
        } else {
            mStorage->wipeDb();
        }
    } else {
        mStorage->wipeDb();
    }

    emit readyToCheckForPendingNotification();
    mChannelService->loadJoinedChannelsFromDb();


}

void RocketChatServerData::setRestApi( RestApi *pNewRestApi )
{

    connect( pNewRestApi, &RestApi::loggedIn, this, &RocketChatServerData::onLoggedIn, Qt::UniqueConnection );
    connect( pNewRestApi, &RestApi::fileDownloaded, this, &RocketChatServerData::onFileDownloaded, Qt::UniqueConnection );
    connect( pNewRestApi, &RestApi::loginError, this, &RocketChatServerData::onLoginError, Qt::UniqueConnection );
    this->mRestApi = pNewRestApi;
    QPair<QString, uint> token = mStorage->getToken();

    if ( token.first.length() && token.second > 0 ) {
        mResumeToken = token.first;
        mTokenExpire = token.second;
    }
}

void RocketChatServerData::loadEmojis()
{

    /* QFile file(":/emoji.json");
     if(!file.open(QFile::ReadOnly)){
         qDebug()<<file.errorString();
         qDebug()<<file.error();
     }

     QJsonDocument doc = QJsonDocument::fromJson(file.readAll());

     QJsonObject obj = doc.object();

     for(QString key: obj.keys()){
         QJsonObject emojiObj = obj[key].toObject();
         QString id = ":"+key+":";
         QString file = "qrc:/res/emojis/"+emojiObj["unicode"].toString()+".png";
         QString html = "<img height='20' width='20' src='"+file+"' />";
         QString unicode = Utils::escapeUnicodeEmoji(emojiObj["unicode"].toString());
         QString category = emojiObj["category"].toString();
         QString sort_order = emojiObj["emoji_order"].toString();
         int sort_orderInt = sort_order.toInt();
         qDebug()<<sort_orderInt;
         mStorage->addCustomEmoji(id,file,html,category,unicode,sort_orderInt);
     }*/


    auto emojiList = mEmojiService->loadEmojisFromDb();

    for ( const auto &emoji : emojiList ) {
        if ( mEmojiRepo != nullptr && !emoji.isNull() ) {
            mEmojiRepo->add( emoji->getIdentifier(), emoji );
            //        if(emoji->getCategory() == "custom"){
            //            mCustomEmojisHash = Utils::hash(mCustomEmojis);
            //        }
        }
    }
}

void RocketChatServerData::switchChannel( const QString &pServer, const QString &pRid, const QString &pName, const QString &pType )
{
    qDebug() << "switch channel to: " << pRid;
    bool switchPossible = false;
    bool readonly = false;

    if ( pRid.length() ) {
        if ( mChannels != nullptr && !mChannels->contains( pRid ) ) {
            if ( pType == "d" || pType == "c" || pType == "p" ) {
                qDebug() << "create new channel object to:" << pRid;
                mChannelService->createChannelObject( pRid, pName, pType );

                if ( mChannels->contains( pRid ) && !mChannels->get( pRid ).isNull() ) {
                    switchPossible = true;
                }
            }
        } else if ( mChannels != nullptr ) {
            auto channel = mChannels->get( pRid );
            readonly = channel->getReadOnly();
            switchPossible = true;
        }
    }

    if ( switchPossible ) {
        qDebug() << "current Channel " << mCurrentChannel;
        qDebug() << "room " << pRid;

        if ( mCurrentChannel != pRid ) {
            emit channelSwitchRequest( pServer, pRid, pName, pType, readonly );
        }
    }
}

void RocketChatServerData::switchChannelByRid( const QString &pName )
{
    if ( mLoggedIn ) {
        if ( mChannels->getChannelByName( pName ) ) {
            auto channel = mChannels->getChannelByName( pName );
            switchChannel( QStringLiteral( "default" ), channel->getRoomId(), pName, channel->getType() );
            mPendingSwitchRoomRequest.clear();
        } else {
            searchForRoomIdByName( pName );
        }
    } else {
        mPendingSwitchRoomRequest = pName;
    }
}

void RocketChatServerData::requestGetChannelDetails( const QString &pChannelId )
{
    QString type;
    QStringList muted;
    bool archived = false;
    bool readonly = false;
    bool selfMuted = false;

    if ( mChannels->contains( pChannelId ) ) {
        auto channel = mChannels->get( pChannelId );

        if ( !channel.isNull() ) {
            archived = channel->getArchived();
            readonly = channel->getReadOnly();
            muted = channel->getMuted();
            selfMuted = channel->getSelfMuted();
            type = channel->getType();



            QVariantList mutedList;

            for ( const auto &entry : muted ) {
                mutedList.append( entry );
            }

            QVariantMap details;

            details[QStringLiteral( "archived" )] = archived;
            details[QStringLiteral( "ro" )] = readonly;
            details[QStringLiteral( "muted" )] = mutedList;
            details[QStringLiteral( "selfMuted" )] = selfMuted;
            details[QStringLiteral( "type" )] = type;
            details[QStringLiteral( "name" )] = channel->getName();
            details[QStringLiteral( "blocked" )] = channel->getBlocked();

            emit channelDetailsReady( details, pChannelId );
        }
    }
}

void RocketChatServerData::requestIsLoggedIn()
{
    if ( mLoggedIn ) {
        emit loggedIn( QStringLiteral( "default" ) );
    }
}

void RocketChatServerData::setUserPresenceStatus( int pStatus )
{
    if ( pStatus > 0 && pStatus < 5 && isWebsocketValid() && diffToLastDDPPing() < 29 ) {

        QSharedPointer<DDPMethodRequest> request( new RocketChatChangeUserPresanceDefaultStatus( static_cast<RocketChatChangeUserPresanceDefaultStatus::status>( pStatus ) ) );
        mDdpApi->sendRequest( request );
    }
}

void RocketChatServerData::onStateChanged( const Qt::ApplicationState &pState )
{
    Q_UNUSED( pState )

    if ( !isWebsocketValid() || diffToLastDDPPing() > 29 ) {
        qDebug() << "call resume";
        resume();
    } else {
        setUserPresenceStatus( 1 );
    }
}

void RocketChatServerData::createVideoCall( const QString &pRid )
{
    if ( !pRid.isEmpty() ) {

        QSharedPointer<DDPMethodRequest> request( new RocketChatUpdateJitsiTimeout( pRid ) );
        mDdpApi->sendRequest( request );
        joinJitsiCall( pRid );
    }
}

//TODO: rename, as method is not really for searching only
void RocketChatServerData::searchForRoomIdByName( const QString &pName )
{
    QSharedPointer<RocketChatSpotLightRequest> request( new RocketChatSpotLightRequest( pName ) );
    std::function<void ( QJsonObject, MeteorDDP * )> success = [ = ]( QJsonObject pResponse, MeteorDDP * ) {

        if ( pResponse.contains( QStringLiteral( "result" ) ) ) {
            QJsonObject result = pResponse[QStringLiteral( "result" )].toObject();

            if ( result.contains( QStringLiteral( "users" ) ) ) {
                QJsonArray users = result[QStringLiteral( "users" )].toArray();

                for ( const auto &user : users ) {
                    QJsonObject userObj = user.toObject();

                    if ( userObj.contains( QStringLiteral( "username" ) ) && userObj.contains( QStringLiteral( "_id" ) ) ) {
                        QString username = userObj[QStringLiteral( "username" )].toString();

                        if ( !username.compare( pName, Qt::CaseInsensitive ) ) {
                            QString id = userObj[QStringLiteral( "_id" )].toString();
                            switchChannel( QStringLiteral( "default" ), id, pName, "d" );
                        }
                    }
                }
            } else if ( result.contains( QStringLiteral( "rooms" ) ) ) {
                QJsonArray rooms = result[QStringLiteral( "rooms" )].toArray();

                for ( const auto &room : rooms ) {
                    QJsonObject roomObj = room.toObject();

                    if ( roomObj.contains( QStringLiteral( "name" ) ) && roomObj.contains( QStringLiteral( "_id" ) ) ) {
                        QString roomname = roomObj[QStringLiteral( "name" )].toString();

                        if ( !roomname.compare( pName, Qt::CaseInsensitive ) ) {
                            QString id = roomObj[QStringLiteral( "_id" )].toString();
                            switchChannel( QStringLiteral( "default" ), id, pName, "c" );
                        }
                    }
                }
            }
        }

        mPendingSwitchRoomRequest.clear();

    };
    request->setSuccess( success );
    mDdpApi->sendRequest( request );
}

void RocketChatServerData::blockUser( const QString &pChannelId )
{
    auto userId = pChannelId;
    userId.replace( mUserId, "" );

    if ( userId == pChannelId ) {
        QString possibleRoom1 = userId + mUserId;
        QString possibleRoom2 = mUserId + userId;

        QSharedPointer<RocketChatBlockUserRequest> blockRequest1( new RocketChatBlockUserRequest( possibleRoom1, userId ) );
        QSharedPointer<RocketChatBlockUserRequest> blockRequest2( new RocketChatBlockUserRequest( possibleRoom2, userId ) );
        mDdpApi->sendRequest( blockRequest1 );
        mDdpApi->sendRequest( blockRequest2 );


    } else {
        QSharedPointer<RocketChatBlockUserRequest> blockRequest( new RocketChatBlockUserRequest( pChannelId, userId ) );
        mDdpApi->sendRequest( blockRequest );
    }
}

void RocketChatServerData::unBlockUser( const QString &pChannelId )
{
    auto userId = pChannelId;
    userId.replace( mUserId, "" );

    if ( userId == pChannelId ) {
        QString possibleRoom1 = userId + mUserId;
        QString possibleRoom2 = mUserId + userId;
        QSharedPointer<RocketChatUnblockUserRequest> unBlockRequest1( new RocketChatUnblockUserRequest( possibleRoom1, userId ) );
        QSharedPointer<RocketChatUnblockUserRequest> unBlockRequest2( new RocketChatUnblockUserRequest( possibleRoom2, userId ) );
        mDdpApi->sendRequest( unBlockRequest1 );
        mDdpApi->sendRequest( unBlockRequest2 );
    } else {
        QSharedPointer<RocketChatUnblockUserRequest> unBlockRequest( new RocketChatUnblockUserRequest( pChannelId, userId ) );
        mDdpApi->sendRequest( unBlockRequest );
    }
}

QString RocketChatServerData::getUserId() const
{
    return mUserId;
}

void RocketChatServerData::setUserId( const QString &userId )
{
    mUserId = userId;
}

void RocketChatServerData::checkForMissedMessages()
{

}

void RocketChatServerData::disconnectFromServer()
{
    mConnectionState == ConnectionState::OFFLINE;
    mDdpApi->disconnectFromServer();
}

void RocketChatServerData::loadHistories()
{
    QStringList channelIds;

    for ( const auto &currentChannel : mChannels->getElements() ) {
        if ( Q_LIKELY( !currentChannel.isNull() && currentChannel->getJoined() ) ) {
            //joinChannel( currentChannel->getRoomId() );
            channelIds.append( currentChannel->getRoomId() );
        }
    }

    QSharedPointer<LoadHistoryServiceRequest> request( new LoadHistoryServiceRequest( channelIds, 1 ) );
    request->setLimit( 1 );
    request->setSuccess( historyLoaded );
    request->setNoLastThreeDays( true );
    mMessageService->loadHistory( request );
}

QString RocketChatServerData::parseMessageReceived( const QString &pMessage )
{
    auto message = Utils::emojiFy( pMessage, mEmojiRepo );
    //    message = Utils::linkiFy( message );
    //    message = message.replace( "\n", "<br>" );
    return message;
}


void RocketChatServerData::onUsersLoaded( const QString &pChannelId, const QVector<QSharedPointer<RocketChatUser> > &pUserList )
{
    if ( Q_LIKELY( mChannels->contains( pChannelId ) ) ) {
        auto channel = mChannels->get( pChannelId );

        if ( !channel.isNull() ) {
            for ( const auto &user : pUserList ) {
                if ( !user.isNull() ) {
                    channel->addUser( user );
                    userModel.insertUser( pChannelId, user );
                }
            }
        }
    }
}

void RocketChatServerData::onUnreadCountChanged()
{
    uint number = 0;
    qDebug() << "on Unread count changed";

    for ( const auto &channel : mChannels->getElements() ) {
        if ( !channel.isNull() ) {
            number += channel->getUnreadMessages();
        }
    }

    emit unreadCountChanged( QStringLiteral( "default" ), number );
}

qint16 RocketChatServerData::getCustomEmojisHash() const
{
    return mCustomEmojisHash;
}

void RocketChatServerData::setCustomEmojisHash( const qint16 &customEmojisHash )
{
    mCustomEmojisHash = customEmojisHash;
}

EmojiRepo *RocketChatServerData::getEmojiRepo() const
{
    return mEmojiRepo;
}

void RocketChatServerData::setEmojiRepo( EmojiRepo *emojiRepo )
{
    mEmojiRepo = emojiRepo;
}

void RocketChatServerData::createAccount( const QString &username, const QString &email, const QString &password )
{
    QSharedPointer<RocketChatCreateAccount> request( new RocketChatCreateAccount( username, email, password ) );

    sendDdprequest( request );
}

void RocketChatServerData::offlineLogin()
{
    QString userId = mStorage->getUserId();
    QString userName = mStorage->getUserName();

    if ( !userId.isEmpty() && !userName.isEmpty() ) {
        this->mUserId = userId;
        emit loggedIn( QStringLiteral( "default" ) );
    }
}

QVariantList RocketChatServerData::getEmojisByCategory() const
{
    QVariantList list;
    auto emojiMap = mEmojiRepo->emojisOrdered();

    for ( const auto &emoji : emojiMap ) {
        list.append( emoji->toQVariantMap() );
    }

    return list;
}

void RocketChatServerData::login( const QString &pUsername, const QString &pPassword )
{
    QPair<QString, uint> tokenDb = mStorage->getToken();
    QString userIdDb = mStorage->getUserId();
    QString userDb = mStorage->getUserName();
    QDateTime currentTime = QDateTime::currentDateTime();

    if ( !tokenDb.first.isEmpty() && tokenDb.second > currentTime.toTime_t() ) {
        loginWithToken( userDb, tokenDb.first );
    } else {
        QString passDb = mStorage->getPassword();
        QString pass;
        QString username = pUsername;

        if ( passDb.length() ) {
            username = userDb;
            pass = passDb;
        } else {
            QByteArray hashArray = QCryptographicHash::hash( pPassword.toUtf8(), QCryptographicHash::Sha256 );
            pass = hashArray.toHex();
        }

        loginWithHash( username, pass );
    }
}

void RocketChatServerData::loginWithHash( const QString &pUsername, const QString &pPswHash )
{
    qDebug() << "login with hash";

    auto self = this;

    QSharedPointer<DDPLoginRequest> request( new DDPLoginRequest( pUsername, pPswHash ) );
    std::function<void ( QJsonObject, MeteorDDP * )> success = [ self, pUsername, pPswHash, request ]( QJsonObject pResponse, MeteorDDP * ) {
        qDebug() << "authenticated";

        if ( pResponse.contains( QStringLiteral( "result" ) ) ) {
            QJsonObject result = pResponse[QStringLiteral( "result" )].toObject();

            if ( result.contains( QStringLiteral( "token" ) ) && result.contains( QStringLiteral( "id" ) ) ) {
                self->mConnectionState = ConnectionState::ONLINE;
                self->mResumeToken = result[QStringLiteral( "token" )].toString();
                QJsonObject expireObject = result[( "tokenExpires" )].toObject();
                double expireDouble = expireObject[QStringLiteral( "$date" )].toDouble();
                self->mTokenExpire = static_cast<uint>( expireDouble / 1000 );
                QString userId = result[QStringLiteral( "id" )].toString();
                self->mStorage->transaction();
                self->mStorage->setUserData( pUsername, pPswHash );
                self->mStorage->setToken( self->mResumeToken, self->mTokenExpire );
                self->mStorage->setUserId( userId );
                self->mRestApi->setToken( self->mResumeToken );
                self->mRestApi->setUserId( userId );
                self->mStorage->commit();
                self->mUserId = result[QStringLiteral( "id" )].toString();
                self->mDdpApi->setToken( self->mResumeToken );
                self->mDdpApi->unsetResponseBinding( request->getFrame() );
                self->onDDPAuthenticated();
                emit( self->onHashLoggedIn( self->mServerId ) );
            }
        }
    };
    DdpCallback error = [ = ]( QJsonObject pResponse, MeteorDDP * ) {
        Q_UNUSED( pResponse );
        onLoginError();
    };

    request->setSuccess( success );
    request->setError( error );
    sendDdprequest( request, true, true );
    this->mUsername = pUsername;
}

void RocketChatServerData::loginWithToken( const QString &pUsername, const QString &pToken, bool pResume )
{
    Q_UNUSED( pResume );
    qDebug() << "login with token";

    qDebug() << "token from db: " << pToken;

    if ( !pToken.isEmpty() ) {
        auto self = this;
        RestRequestCallback meCallBackSuccess = [ self ]( QNetworkReply *, QJsonObject data, RestApi * ) {

            if ( data.contains( QStringLiteral( "username" ) ) ) {
                self->mUsername = data[QStringLiteral( "username" )].toString();
                self->mStorage->setUserName( self->mUsername );
                self->onResume();
                self->onDDPAuthenticated();
            }

        };

        QSharedPointer<DDPLoginRequest> request( new DDPLoginRequest( pToken ) );
        std::function<void ( QJsonObject, MeteorDDP * )> success = [ self, request, meCallBackSuccess ]( QJsonObject pResponse, MeteorDDP * ) {
            qDebug() << "authenticated";
            self->mConnectionState = ConnectionState::ONLINE;

            if ( pResponse.contains( QStringLiteral( "result" ) ) ) {
                QJsonObject result = pResponse[QStringLiteral( "result" )].toObject();

                if ( result.contains( QStringLiteral( "token" ) ) && result.contains( QStringLiteral( "id" ) ) ) {
                    self->mResumeToken = result[QStringLiteral( "token" )].toString();
                    QJsonObject expireObject = result[QStringLiteral( "tokenExpires" )].toObject();
                    double expireDouble = expireObject[QStringLiteral( "$date" )].toDouble();
                    self->mTokenExpire = static_cast<uint>( expireDouble / 1000 );

                    QString userId = result[QStringLiteral( "id" )].toString();
                    self->mStorage->setToken( self->mResumeToken, self->mTokenExpire );
                    self->mStorage->setUserId( userId );
                    self->mRestApi->setToken( self->mResumeToken );
                    self->mRestApi->setUserId( userId );
                    self->mUserId = result[QStringLiteral( "id" )].toString();
                    self->mDdpApi->setToken( self-> mResumeToken );
                    self->mDdpApi->unsetResponseBinding( request->getFrame() );

                    RestApiRequest meRequest
                        = RestApiRequest( new restMeRequest( meCallBackSuccess ) );

                    self->mRestApi->sendRequest( meRequest );
                }
            }
        };
        DdpCallback error = [ = ]( QJsonObject pResponse, MeteorDDP * ) {
            Q_UNUSED( pResponse );
            qDebug() << "login error";
            //onLoginError();
            //try again with hash

            QString userDb = mStorage->getUserName();
            QString passDb = mStorage->getPassword();

            if ( !userDb.isEmpty() && !passDb.isEmpty() ) {

                loginWithHash( userDb, passDb );
            } else {
                qWarning() << "token rejected and no user+hash information present";
                onLoginError();
            }
        };

        request->setSuccess( success );
        request->setError( error );
#ifdef Q_OS_IOS
    mDdpApi->sendRequest(request);
#else
        sendDdprequest( request, true, true );
#endif
        this->mUsername = pUsername;
    } else {
        qDebug() << "empty token";
    }
}

void RocketChatServerData::loginWtihSamlToken( const QString &pToken )
{
    if ( !pToken.isEmpty() ) {
        qDebug() << pToken;
        
        RestRequestCallback meCallBackSuccess = [ = ]( QNetworkReply *, QJsonObject data, RestApi * ) {

            if ( data.contains( QStringLiteral( "username" ) ) ) {
                mUsername = data[QStringLiteral( "username" )].toString();
                mStorage->setUserName( mUsername );
                onResume();
                this->onDDPAuthenticated();
            }
            
        };
        
        QSharedPointer<ddpSamlLoginRequest> request( new ddpSamlLoginRequest( pToken ) );
        std::function<void ( QJsonObject, MeteorDDP * )> success = [ = ]( QJsonObject pResponse, MeteorDDP * ) {
            qDebug() << "authenticated";
            mConnectionState = ConnectionState::ONLINE;
            
            if ( pResponse.contains( QStringLiteral( "result" ) ) ) {
                QJsonObject result = pResponse[QStringLiteral( "result" )].toObject();
                
                if ( result.contains( QStringLiteral( "token" ) ) && result.contains( QStringLiteral( "id" ) ) ) {
                    this->mResumeToken = result[QStringLiteral( "token" )].toString();
                    QJsonObject expireObject = result[QStringLiteral( "tokenExpires" )].toObject();
                    double expireDouble = expireObject[QStringLiteral( "$date" )].toDouble();
                    mTokenExpire = static_cast<uint>( expireDouble / 1000 );
                    
                    QString userId = result[QStringLiteral( "id" )].toString();
                    mStorage->setToken( mResumeToken, mTokenExpire );
                    mStorage->setUserId( userId );
                    mRestApi->setToken( mResumeToken );
                    mRestApi->setUserId( userId );
                    this->mUserId = result[QStringLiteral( "id" )].toString();
                    mDdpApi->setToken( mResumeToken );
                    mDdpApi->unsetResponseBinding( request->getFrame() );
                    
                    
                    RestApiRequest meRequest
                        = RestApiRequest( new restMeRequest( meCallBackSuccess ) );

                    mRestApi->sendRequest( meRequest );
                    
                    
                    
                }
            }
        };
        DdpCallback error = [ = ]( QJsonObject pResponse, MeteorDDP * ) {
            Q_UNUSED( pResponse );
            qDebug() << "login error";
            qWarning() << "SAML2 token rejected";
            onLoginError();
        };
        
        request->setSuccess( success );
        request->setError( error );
        sendDdprequest( request, true, true );
    } else {
        qDebug() << "empty token";
    }
}

void RocketChatServerData::loginWithOpenIDToken( const QString &pToken, const QString &pSecret )
{
    if ( !pToken.isEmpty() && !pSecret.isEmpty() ) {
        auto self = this;
        RestRequestCallback meCallBackSuccess = [ self ]( QNetworkReply *, QJsonObject data, RestApi * ) {

            if ( data.contains( QStringLiteral( "username" ) ) ) {
                self->mUsername = data[QStringLiteral( "username" )].toString();
                self->mStorage->transaction();
                self->mStorage->setUserData( self->mUsername, "" );
                self->mStorage->setToken( self->mResumeToken, self->mTokenExpire );
                self->mStorage->setUserId( self->mUserId );
                self->mStorage->commit();
                self->onResume();
                self->onDDPAuthenticated();
            }

        };

        QSharedPointer<DDPOpenIDLoginRequest> request( new DDPOpenIDLoginRequest( pToken, pSecret ) );
        std::function<void ( QJsonObject, MeteorDDP * )> success = [ = ]( QJsonObject pResponse, MeteorDDP * ) {
            qDebug() << "authenticated";
            self->mConnectionState = ConnectionState::ONLINE;

            if ( pResponse.contains( QStringLiteral( "result" ) ) ) {
                QJsonObject result = pResponse[QStringLiteral( "result" )].toObject();

                if ( result.contains( QStringLiteral( "token" ) ) && result.contains( QStringLiteral( "id" ) ) ) {
                    self->mConnectionState = ConnectionState::ONLINE;
                    self->mResumeToken = result[QStringLiteral( "token" )].toString();
                    QJsonObject expireObject = result[QStringLiteral( "tokenExpires" )].toObject();
                    double expireDouble = expireObject[QStringLiteral( "$date" )].toDouble();
                    self->mTokenExpire = static_cast<uint>( expireDouble / 1000 );

                    QString userId = result[QStringLiteral( "id" )].toString();

                    self->mRestApi->setToken( mResumeToken );
                    self->mRestApi->setUserId( userId );
                    self->mUserId = result[QStringLiteral( "id" )].toString();
                    self->mDdpApi->setToken( mResumeToken );
                    self->mDdpApi->unsetResponseBinding( request->getFrame() );

                    emit self->onHashLoggedIn( self->mServerId ) ;

                    RestApiRequest meRequest
                        = RestApiRequest( new restMeRequest( meCallBackSuccess ) );

                    self->mRestApi->sendRequest( meRequest );

                }
            }
        };
        DdpCallback error = [ = ]( QJsonObject pResponse, MeteorDDP * ) {
            Q_UNUSED( pResponse );
            qDebug() << "login error";

            qWarning() << "token rejected and no user+hash information present";
            onLoginError();
        };

        request->setSuccess( success );
        request->setError( error );
        sendDdprequest( request, true, true );

    } else {
        qDebug() << "empty token";
    }
}

void RocketChatServerData::loginWithMethod( const QString &method, const QString &payload )
{
    if ( mLoginMethodRepo.contains( method ) ) {
        LoginMethod loginMethod = mLoginMethodRepo[method];

        if ( loginMethod.getType() == LoginMethod::Type::OPENID ) {
            QString token;
            QString secret;
            QJsonDocument doc = QJsonDocument::fromJson( payload.toUtf8() );
            QJsonObject obj = doc.object();

            if ( obj.contains( QStringLiteral( "credentialToken" ) ) && obj.contains( QStringLiteral( "credentialSecret" ) ) ) {
                token = obj[QStringLiteral( "credentialToken" )].toString();
                secret = obj[QStringLiteral( "credentialSecret" )].toString();
                // if app went offline during webview
                disconnectFromServer();
                loginWithOpenIDToken( token, secret );
                resume();
            }
        }
    }
}

void RocketChatServerData::onDDPConnected()
{
    QSharedPointer<PingRequest> request( new PingRequest() );
    mDdpApi->sendRequest( request );

    if ( mResumeOperation ) {
        loginWithToken( mUsername, mDdpApi->getToken(), true );
    } else {
        QPair<QString, uint> tokenDb = mStorage->getToken();
        QString token = tokenDb.first;
        QString userDb = mStorage->getUserName();

        if ( !token.isEmpty() && !userDb.isEmpty() ) {
            loginWithToken( userDb, token );
        }
    }

    mConnectionState = ConnectionState::CONNECTED;
    getServerSettings();
    getLoginMethods();
    emit( ddpConnected( mServerId ) );
    mConnected = true;
    sendUnsentMessages();
}

void RocketChatServerData::onDDPDisonnected()
{
    mConnectionState = ConnectionState::OFFLINE;
}

void RocketChatServerData::onDDPAuthenticated()
{
    //TODO: dirty fix!
    getCustomEmojis();
    setUserPresenceStatus( 1 );
    mStorage->setSetting( QStringLiteral( "currentServer" ), mBaseUrl );
    getServerInfo();

    if ( !mCurrentChannel.isEmpty() ) {
        joinChannel( mCurrentChannel );
    }

}

void RocketChatServerData::handleChannelMessage( const QJsonObject &pMessage )
{
    if ( Q_LIKELY( pMessage.contains( QStringLiteral( "fields" ) ) ) ) {
        QJsonObject fields = pMessage[QStringLiteral( "fields" )].toObject();

        if ( Q_LIKELY( fields.contains( QStringLiteral( "args" ) ) ) ) {
            QJsonArray args = fields[QStringLiteral( "args" )].toArray();
            QString firstArg = args[0].toString();

            if ( firstArg == QStringLiteral( "insert" ) || firstArg == QStringLiteral( "inserted" ) ) {
                QJsonObject newChannel = args[1].toObject();

                if ( newChannel.contains( QStringLiteral( "name" ) ) && newChannel.contains( QStringLiteral( "rid" ) ) && newChannel.contains( "t" ) ) {
                    QString name = newChannel[QStringLiteral( "name" )].toString();
                    QString rid = newChannel[QStringLiteral( "rid" )].toString();

                    if ( !mChannels->contains( QStringLiteral( "id" ) ) ) {
                        QSharedPointer<RocketChatChannel> channel  = mChannelService->createChannelObject( rid, name, newChannel["t"].toString() );

                        if ( !channel.isNull() ) {
                            channel->setJoined( true );

                            //TODO: check
                            if ( mChannels->add( rid, channel ) ) {

                            }
                        }
                    }
                }
            } else if ( firstArg == QStringLiteral( "updated" ) ) {
                QJsonObject data = args[1].toObject();
                int unread = data[QStringLiteral( "unread" )].toInt();
                QString rid = data[QStringLiteral( "rid" )].toString();
                bool blocked = false;
                QString name = data[QStringLiteral( "name" )].toString();
                bool hidden = false;

                if ( data.contains( QStringLiteral( "open" ) ) ) {
                    hidden = !data[QStringLiteral( "open" )].toBool();
                }

                if ( data.contains( QStringLiteral( "blocker" ) ) ) {
                    blocked = data[QStringLiteral( "blocker" )].toBool();

                    //TODO: clean this up!, there should be a user service
                    QString otherUserId = rid;
                    otherUserId = otherUserId.replace( mUserId, "" );

                    if ( data.contains( "name" ) ) {
                        QString otherUserName = data["name"].toString();
                        mStorage->addUserToBlockList( otherUserName, otherUserId );
                        mMessageService->addUserToBlockList( otherUserId );
                        leaveChannel( rid );
                    }
                }

                if ( hidden ) {
                    mChannelService->deleteChannel( rid );
                } else {
                    if ( mChannels->contains( rid ) && !mChannels->get( rid ).isNull() ) {
                        auto channel =  mChannels->get( rid );
                        channel->setName( name );
                        channel->setBlocked( blocked );
                        channel->setUnreadMessages( unread );

                    } else {
                        QSharedPointer<RocketChatChannel> channel  = mChannelService->createChannelObject( rid, name, data["t"].toString() );
                        channel->setBlocked( blocked );
                        channel->setUnreadMessages( unread );
                        mChannels->add( rid, channel );

                    }
                }
            }
        }
    }
}

uint RocketChatServerData::diffToLastDDPPing()
{
    return mDdpApi->diffToLastMessage();
}

void RocketChatServerData::getFileRessource( const QString &pUrl )
{
    getFileRessource( pUrl, QStringLiteral( "temp" ) );
}

void RocketChatServerData::getFileRessource( const QString &pUrl, const QString &pType )
{
    bool showInline = false;

    if ( pType == QStringLiteral( "temp" ) ) {
        showInline = true;
    }

    auto then = [ = ]( QSharedPointer<TempFile> tempfile, bool showInline ) {
        emit( fileRessourceProcessed( pUrl, tempfile->getFilePath(), showInline ) );
    };
    QSharedPointer< FileRequest > request( new FileRequest( pUrl, pType, then, showInline ) );
    mFileService->getFileRessource( request );

}

void RocketChatServerData::sendPushToken( const QString &pToken )
{
    if ( !mUserId.isEmpty() && !pToken.isEmpty() ) {
        qDebug() << "send push token " + pToken;
        QSharedPointer<DDPRequest> sendToken = QSharedPointer<DDPRequest>( new RocketChatUpdatePushTokenRequest( pToken, mUserId ) );
        mDdpApi->sendRequest( sendToken );

    }
}


void RocketChatServerData::requestAllChannels()
{
    QVariantList channelsMap;

    for ( const auto &channel : mChannels->getElements() ) {
        if ( !channel.isNull() ) {
            QVariantMap obj;
            obj[QStringLiteral( "name" )] = channel->getName();
            obj[QStringLiteral( "id" )] = channel->getRoomId();
            channelsMap.append( obj );
        }
    }

    emit allChannelsReady( channelsMap );
}

void RocketChatServerData::requestUsersOfChannel( const QString &pChannelId )
{
    auto users = getUserOfChannel( pChannelId );
    QJsonArray array;

    for ( const auto &current : users ) {
        if ( !current.isNull() ) {
            array.append( current->getUserName() );
        }
    }

    //TODO: use QVariant...
    QJsonDocument doc( array );
    emit usersReady( doc.toJson(), pChannelId );
}

void RocketChatServerData::sendUnsentMessages()
{
    qDebug() << "send unsend ddp messages";

    if ( mConnectionState == ConnectionState::ONLINE
       ) {
        mUnsentMutex.lock();

        for ( const auto &request : mUnsendDdpRequests ) {
            if ( !request.isNull() ) {
                sendDdprequest( request );
            }
        }

        mUnsendDdpRequests.clear();
        qDebug() << "send unsend rest messages";

        for ( const auto &request : mUnsendRestRequests ) {
            if ( !request.isNull() ) {
                sendApiRequest( request );
            }
        }

        mUnsendRestRequests.clear();
        mUnsentMutex.unlock();
        qDebug() << "all messages send";
    } else if ( mConnectionState == ConnectionState::CONNECTED ) {
        mUnsentMutex.lock();

        for ( const auto &request : mUnsendDdpRequestsOnConnected ) {
            if ( !request.isNull() ) {
                sendDdprequest( request, true, true );
            }
        }

        mUnsendDdpRequestsOnConnected.clear();
        qDebug() << "send unsend rest messages";

        mUnsentMutex.unlock();
    }
}

void RocketChatServerData::sendApiRequest( const RestApiRequest &pRequest, bool pRetry )
{
    RestRequestCallback originalCallback = pRequest->getError();

    if ( pRetry ) {
        RestRequestCallback errorWrapperFunction = [ = ]( QNetworkReply * pReply, QJsonObject pData, RestApi * pRestApi ) {
            if ( pReply && pRestApi ) {
                mUnsendRestRequests.append( pRequest );

                if ( originalCallback ) {
                    originalCallback( pReply, pData, pRestApi );
                }
            }
        };
        pRequest->setError( errorWrapperFunction );
    }

    mRestApi->sendRequest( pRequest );
}

void RocketChatServerData::sendDdprequest( const QSharedPointer<DDPRequest> &pRequest, bool pRetry, bool pSendWhenConnected )
{
    if ( !pRequest.isNull() ) {
        uint diff = diffToLastDDPPing();
        qDebug( "%d diff to last ping %d", static_cast<int>( mConnectionState ), diff );
        bool intime = diff <= 29;

        if ( Q_LIKELY( mConnectionState == ConnectionState::ONLINE && intime ) ) {
            mDdpApi->sendRequest( pRequest );
        } else if ( mConnectionState == ConnectionState::CONNECTED && intime && pSendWhenConnected ) {
            mDdpApi->sendRequest( pRequest );
        } else if ( pRetry && pSendWhenConnected ) {
            mUnsendDdpRequestsOnConnected.append( pRequest );
        } else if ( pRetry ) {
            qDebug() << "message queued";
            mUnsendDdpRequests.append( pRequest );
        }
    }
}

void RocketChatServerData::onDDPMessageReceived( const QJsonObject &pMessage )
{
    sendUnsentMessages();

    if ( Q_LIKELY( pMessage.contains( QStringLiteral( "collection" ) ) ) ) {

        if ( pMessage[QStringLiteral( "collection" )] == QStringLiteral( "stream-notify-user" ) ) {
            handleChannelMessage( pMessage );
        } else if ( pMessage[QStringLiteral( "collection" )] == QStringLiteral( "users" ) ) {
            QString userId = pMessage[QStringLiteral( "id" )].toString();

            if ( mUsers.contains( userId ) ) {
                QString msg = pMessage[QStringLiteral( "msg" )].toString();

                if ( msg == QStringLiteral( "added" ) ) {
                    mUsers.get( userId )->setStatus( RocketChatUser::status::ONLINE );
                } else if ( msg == QStringLiteral( "changed" ) ) {
                    QJsonObject fields = pMessage["fields"].toObject();

                    if ( fields.contains( QStringLiteral( "status" ) ) ) {
                        QString msgStatus = fields[QStringLiteral( "status" )].toString();
                        RocketChatUser::status userStatus;

                        if ( msgStatus == QStringLiteral( "online" ) ) {
                            userStatus = RocketChatUser::status::ONLINE;
                        } else if ( msgStatus == QStringLiteral( "away" ) ) {
                            userStatus = RocketChatUser::status::AWAY;
                        } else {
                            userStatus = RocketChatUser::status::OFFLINE;
                        }

                        mUsers.get( userId )->setStatus( userStatus );
                    }

                } else if ( msg == QStringLiteral( "removed" ) ) {
                    mUsers.get( userId )->setStatus( RocketChatUser::status::OFFLINE );
                }
            } else {

            }
        } else if ( pMessage[QStringLiteral( "collection" )] == QStringLiteral( "autocompleteRecords" ) ) {
            if ( Q_LIKELY( pMessage.contains( QStringLiteral( "fields" ) ) && pMessage.contains( QStringLiteral( "msg" ) ) && pMessage.contains( QStringLiteral( "id" ) ) ) ) {
                QJsonObject fields = pMessage[QStringLiteral( "fields" )].toObject();

                if ( pMessage[QStringLiteral( "msg" )] == QStringLiteral( "added" ) ) {
                    mAutocompleteRecords[pMessage[QStringLiteral( "id" )].toString()] = fields;
                } else if ( pMessage[QStringLiteral( "msg" )] == QStringLiteral( "removed" ) ) {
                    mAutocompleteRecords.remove( pMessage[QStringLiteral( "id" )].toString() );
                }
            }
        } else if ( pMessage[QStringLiteral( "collection" )] == QStringLiteral( "meteor_accounts_loginServiceConfiguration" ) ) {
            if ( Q_LIKELY( pMessage.contains( QStringLiteral( "fields" ) ) && pMessage.contains( QStringLiteral( "msg" ) ) && pMessage.contains( QStringLiteral( "id" ) ) ) ) {
                QString msg = pMessage[QStringLiteral( "msg" )].toString();

                if ( msg == QStringLiteral( "added" ) ) {
                    QJsonObject data = pMessage[QStringLiteral( "fields" )].toObject();

                    if ( data.contains( QStringLiteral( "scope" ) ) ) {
                        QString scope = data[QStringLiteral( "scope" )].toString();

                        if ( scope == QStringLiteral( "openid" ) ) {
                            // QMap<QString, QVariant> entry = data.toVariantMap();
                            QString service = data[QStringLiteral( "service" )].toString();
                            mLoginMethodRepo.addLoginMethod( data, QStringLiteral( "https://" ) + mBaseUrl );

                            if ( mLoginMethodRepo.contains( service ) ) {
                                LoginMethod loginMethod = mLoginMethodRepo.get( service );
                                emit newLoginMethod( loginMethod );
                            }
                        }
                    }
                }
            }
        } else if ( isUserJoinMessage( pMessage ) ) {
            handleUserJoinMessage( pMessage );
        } else if ( Q_LIKELY( isStreamRoomMessage( pMessage ) ) ) {
            handleStreamRoomMessage( pMessage );
        }
    }
}

bool RocketChatServerData::isStreamRoomMessage( const QJsonObject &pMessage ) const
{
    if ( Q_LIKELY( pMessage.contains( QStringLiteral( "collection" ) ) && pMessage[QStringLiteral( "collection" )] == QStringLiteral( "stream-room-messages" ) ) ) {
        return true;
    }

    return false;
}

bool RocketChatServerData::isUserJoinMessage( const QJsonObject &pMessage ) const
{
    if ( pMessage.contains( QStringLiteral( "fields" ) ) && pMessage[QStringLiteral( "fields" )].toObject().contains( "t" ) && pMessage[QStringLiteral( "fields" )].toObject()["t"] == QStringLiteral( "uj" ) ) {
        return true;
    }

    return false;
}

void RocketChatServerData::handleStreamRoomMessage( const QJsonObject &pMessage )
{
    if ( Q_LIKELY( pMessage.contains( QStringLiteral( "fields" ) ) ) ) {
        QJsonObject fields =  pMessage[QStringLiteral( "fields" )].toObject();

        if ( Q_LIKELY( fields.contains( QStringLiteral( "args" ) ) ) ) {
            QJsonArray args = fields[QStringLiteral( "args" )].toArray();

            for ( const auto &currentArg : args ) {
                QJsonObject currentArgObj = currentArg.toObject();

                if ( Q_LIKELY( currentArgObj.contains( QStringLiteral( "rid" ) ) && currentArgObj.contains( QStringLiteral( "_id" ) ) ) ) {
                    QString roomId = currentArgObj[QStringLiteral( "rid" )].toString();
                    auto message = mMessageService->parseMessage( currentArgObj, true );

                    if ( Q_LIKELY( mChannels->contains( roomId ) ) ) {
                        mChannels->get( roomId )->addMessage( message );
                        mMessageService->persistMessage( message );
                    }
                }
            }
        }
    }
}

void RocketChatServerData::handleUserJoinMessage( const QJsonObject &pMessage )
{
    if ( Q_LIKELY( pMessage.contains( QStringLiteral( "fields" ) ) ) ) {
        QJsonObject fields  = pMessage[QStringLiteral( "fields" )].toObject();

        if ( Q_LIKELY( fields.contains( "rid" ) && fields.contains( "u" ) ) ) {
            QString roomId = fields[QStringLiteral( "rid" )].toString();
            QJsonObject userData = fields["u"].toObject();

            if ( Q_LIKELY( userData.contains( QStringLiteral( "_id" ) ) && userData.contains( QStringLiteral( "username" ) ) ) ) {
                QString userId = userData[QStringLiteral( "_id" )].toString();
                QString username = userData[QStringLiteral( "username" )].toString();

                if ( Q_LIKELY( mChannels->contains( roomId ) ) ) {
                    QSharedPointer<RocketChatChannel> channel = mChannels->get( roomId );

                    if ( !channel.isNull() ) {
                        QSharedPointer<RocketChatUser> ptr( new RocketChatUser( username ) );
                        ptr->setUserId( userId );

                        if ( Q_LIKELY( channel->addUser( ptr ) ) ) {
                            emit userJoined( roomId, username );
                        }
                    }
                }
            }
        }
    }
}

void RocketChatServerData::onLoginError()
{
    qDebug() << "login error";
    //if there is a login error reset anything
    mConnectionState = ConnectionState::CONNECTED;
    emit loginError();
}

void RocketChatServerData::onResume()
{
    qDebug() << "resubscribing channel";
    std::tuple<QString, QString> channelData = mStorage->getCurrentChannel();
    QString channelId = std::get<0>( channelData );
    sendUnsentMessages();

    if ( mChannels->contains( channelId ) ) {
        auto channel = mChannels->get( channelId );
        mChannelService->subscribeChannel( channel );
        //loadMissedMessages();
        qDebug() << "connection resumed";

    }

    if ( !mCurrentChannel.isEmpty() && mChannels->contains( mCurrentChannel ) ) {
        mChannels->get( mCurrentChannel )->setUnreadMessages( 0 );
    }

}


QJsonObject RocketChatServerData::formatMessageTime( const QJsonObject &pJsonObject )
{
    if ( Q_LIKELY( pJsonObject.contains( QStringLiteral( "ts" ) ) ) ) {
        QJsonObject ts = pJsonObject[QStringLiteral( "ts" )].toObject();

        if ( Q_LIKELY( ts.contains( QStringLiteral( "$date" ) ) ) ) {
            double timeStamp = ts[QStringLiteral( "$date" )].toDouble();
            QDateTime dateTime = QDateTime::fromMSecsSinceEpoch( timeStamp );
            QString formattedTime = dateTime.toString( QStringLiteral( "hh:mm" ) );
            QString formattedDate = dateTime.toString( QStringLiteral( "dd.MM.yyyy" ) );
            pJsonObject[QStringLiteral( "formattedDate" )] = formattedDate;
            pJsonObject[QStringLiteral( "formattedTime" )] = formattedTime;
            pJsonObject[QStringLiteral( "msg" )] = parseMessageReceived( pJsonObject[QStringLiteral( "msg" )].toString() );
            return pJsonObject;
        }
    }

    return QJsonObject();
}

void RocketChatServerData::joinJitsiCall( const QString &pRoomId )
{

    if ( Q_LIKELY( mChannels->contains( pRoomId ) ) ) {
        QString hash;
        bool defaultHash = true;

        int major = std::get<0>( RocketChatServerConfig::serverVersion );
        int minor = std::get<1>( RocketChatServerConfig::serverVersion );
        int patch = std::get<2>( RocketChatServerConfig::serverVersion );

        if ( major != -1 && minor != -1 && patch != -1 ) {
            if ( major >= 0 && minor > 60 ) {
                hash = "undefined" + pRoomId;
                defaultHash = false;
            }
        }

        if ( defaultHash ) {
            hash = QCryptographicHash::hash( ( RocketChatServerConfig::uniqueId + pRoomId ).toUtf8(), QCryptographicHash::Md5 ).toHex();
        }

#if defined(Q_OS_IOS) || defined(Q_OS_ANDROID)
        QString scheme = QStringLiteral( "org.jitsi.meet://" );
#else
        QString scheme = "https://";
#endif
        QString url = scheme + RocketChatServerConfig::jitsiMeetUrl + '/' + RocketChatServerConfig::jitsiMeetPrefix + hash;
        qDebug() << url;
        emit openUrl( url );
    }
}

void RocketChatServerData::setUnreadMessages( const QString &pChannelId, int count )
{
    Q_UNUSED( count )

    if ( mChannels->contains( pChannelId ) ) {
        auto channel = mChannels->get( pChannelId );
        channel->setUnreadMessages( 0 );
    }
}

ConnectionState RocketChatServerData::getConnectionState() const
{
    return mConnectionState;
}

void RocketChatServerData::setConnectionState( const ConnectionState &pValue )
{
    mConnectionState = pValue;
}

QString RocketChatServerData::getCurrentChannel() const
{
    return mCurrentChannel;
}

void RocketChatServerData::setCurrentChannel( const QString &pCurrentChannel )
{
    if ( mChannels->contains( pCurrentChannel ) && !mChannels->get( pCurrentChannel ).isNull() ) {
        mCurrentChannel = pCurrentChannel;
        auto channel = mChannels->get( pCurrentChannel );
        mStorage->setCurrentChannel( pCurrentChannel, channel->getName() );
    } else if ( pCurrentChannel == "" ) {
        mCurrentChannel = pCurrentChannel;
        mStorage->setCurrentChannel( "", "" );
    }
}

QString RocketChatServerData::getUsername() const
{
    return mUsername;
}

void RocketChatServerData::setUsername( const QString &pUsername )
{
    mUsername = pUsername;
    mStorage->setUserName( pUsername );
}

void RocketChatServerData::getServerSettings()
{
    QJsonArray params = {};
    QSharedPointer<DDPMethodRequest> request( new DDPMethodRequest( QStringLiteral( "public-settings/get" ), params ) );
    DdpCallback sucess = [ = ]( QJsonObject data, MeteorDDP * ddpApi ) {
        Q_UNUSED( ddpApi );
        QJsonArray configs = data[QStringLiteral( "result" )].toArray();

        for ( const auto &currentConfig : configs ) {
            QJsonObject currentConfObject = currentConfig.toObject();
            QString id    = currentConfObject[QStringLiteral( "_id" )].toString();
            QString value = currentConfObject[QStringLiteral( "value" )].toString();

            if ( id  == QStringLiteral( "uniqueID" ) ) {
                RocketChatServerConfig::uniqueId = value;
            }

            if ( id == QStringLiteral( "Jitsi_Domain" ) ) {
                RocketChatServerConfig::jitsiMeetUrl = value;
            }

            if ( id == QStringLiteral( "Jitsi_URL_Room_Prefix" ) ) {
                RocketChatServerConfig::jitsiMeetPrefix = value;
            }
        }

        qDebug() << "processed config";
    };
    request->setSuccess( sucess );
    sendDdprequest( request, 1, 1 );
}

void RocketChatServerData::onChannelsLoaded( const QVector<QSharedPointer<RocketChatChannel>> &pChannels )
{

    qDebug() << "on channels loaded ";
    loadHistories();
}

void RocketChatServerData::onFileDownloaded( const QString &pUrl, const QString &pTempfile )
{
    Q_UNUSED( pUrl );
    Q_UNUSED( pTempfile );
}

void RocketChatServerData::onLoggedIn()
{
    mLoggedIn = true;
    mChannelService->loadJoinedChannelsFromServer();
    emit loggedIn( mServerId );

    if ( !mPendingSwitchRoomRequest.isEmpty() ) {
        switchChannelByRid( mPendingSwitchRoomRequest );
    }
}

void RocketChatServerData::markChannelAsRead( const QString &pChannelId )
{
    if ( Q_LIKELY( mChannels->contains( pChannelId ) ) ) {
        QSharedPointer<DDPRequest> markAsRead( new RocketChatMarkChannelAsReadRequest( pChannelId ) );
        mDdpApi->sendRequest( markAsRead );
    }
}
void RocketChatServerData::joinChannel( const QString &pChannelId )
{
    if ( Q_LIKELY( mChannels->contains( pChannelId ) ) ) {
        //  if(mCurrentChannel != pChannelId){
        mCurrentChannel = pChannelId;
        auto channel = mChannels->get( pChannelId );
        mChannelService->subscribeChannel( channel );
        setUnreadMessages( pChannelId, 0 );
        markChannelAsRead( pChannelId );
        setCurrentChannel( pChannelId );
        QSharedPointer<LoadHistoryServiceRequest> serviceRequest( new LoadHistoryServiceRequest( pChannelId ) );
        serviceRequest->setSuccess( historyLoaded );
        serviceRequest->setLimit( 300 );
        mMessageService->loadHistory( serviceRequest );
        // }
    } else {
        qCritical() << "Subscribing channel that is not existent.";
    }
}


void RocketChatServerData::connectWithServer()
{
    mDdpApi->connectWithServer();
}

QString RocketChatServerData::getServerId() const
{
    return mServerId;
}

void RocketChatServerData::setServerId( const QString &pValue )
{
    mServerId = pValue;
}

void RocketChatServerData::setServerAddress( const QString &pServerAddress )
{
    mBaseUrl = pServerAddress;
    mApiUri = QStringLiteral( "https://" ) + pServerAddress + QStringLiteral( "/api/v1" );
}

bool RocketChatServerData::isConnected() const
{
    return mConnected;
}

bool RocketChatServerData::isWebsocketValid() const
{
    return mDdpApi->websocketIsValid();
}

void RocketChatServerData::loadRecentHistory( const QString &pChannelId )
{
    qDebug() << "load recent history called";

    if ( Q_LIKELY( mChannels->contains( pChannelId ) ) ) {
        auto channel = mChannels->get( pChannelId );
        auto oldestMessage = channel->getOldestMessage();
        double end = QDateTime::currentMSecsSinceEpoch();

        if ( !oldestMessage.isNull() ) {
            end = oldestMessage->getTimestamp();
        }

        double start = end - ( 86400 * 3 * 1000 );
        QSharedPointer<LoadHistoryServiceRequest> request( new LoadHistoryServiceRequest( pChannelId ) );
        request->setNoLastThreeDays( true );
        request->setStart( start );
        request->setEnd( end );
        std::function<void ( QMultiMap<QString, QSharedPointer<RocketChatMessage> > *messages )> success =
        [ = ]( QMultiMap<QString, QSharedPointer<RocketChatMessage> > *messages ) {
            if ( messages ) {
                auto messageList = messages->values( pChannelId );
                qDebug() << "load recent history successfull loaded " << messageList.size();
                mChannels->get( pChannelId )->addMessages( messageList );
                mMessageService->persistMessages( messageList );

                if ( messages != nullptr ) {
                    delete messages;
                }
            }
        };
        request->setSuccess( success );
        mMessageService->loadHistory( request );
    }
}

void RocketChatServerData::sendMessage( const QString &pMessage, const QString &pChannelId )
{
    ChatMessage message( new RocketChatMessage( pChannelId, pMessage ) );
    mMessageService->sendMessage( message );
}


void RocketChatServerData::getCustomEmojis()
{
    std::function<void ( QList<QSharedPointer<Emoji>> )> success = [ = ]( QList<QSharedPointer<Emoji>> pEmojiList ) {
        QVariantList variantList;

        if ( variantList.size() ) {
            mStorage->transaction();

            for ( const auto &emoji : pEmojiList ) {
                if ( !emoji.isNull() ) {
                    mEmojiService->persistEmoji( emoji );

                    customEmojisDirtyFlag = true;

                    if ( !mEmojiRepo->contains( emoji->getIdentifier() ) ) {

                        mEmojiRepo->add( emoji->getIdentifier(), emoji );
                        variantList.append( emoji->toQVariantMap() );

                        if ( emoji->getCategory() == QStringLiteral( "custom" ) ) {
                            // mCustomEmojisHash = Utils::hash(mCustomEmojis);
                        }
                    }
                }
            }

            mStorage->commit();
        }

        emit customEmojisReceived( getEmojisByCategory() );


        QSharedPointer<RocketChatSubscribeUserNotify> notifySubSubscription( new RocketChatSubscribeUserNotify( this->mUserId ) );
        QSharedPointer<RocketChatNotifyNoticesRequest> notifyNoticeSubSubscription( new RocketChatNotifyNoticesRequest( this->mUserId ) );
        QSharedPointer<RocketChatSubScriptionChangedRequest> noitfySubscriptionsSubscription( new RocketChatSubScriptionChangedRequest( this->mUserId ) );
        QSharedPointer<RocketChatSubscribeActiveUsers> activeUsersSubscription( new RocketChatSubscribeActiveUsers() );
        mDdpApi->sendRequest( notifySubSubscription );
        mDdpApi->sendRequest( notifyNoticeSubSubscription );
        mDdpApi->sendRequest( noitfySubscriptionsSubscription );
        mDdpApi->sendRequest( activeUsersSubscription );

        onLoggedIn();
        sendUnsentMessages();
        getServerSettings();
        emit registerForPush();

        qDebug() << "ddp authenticated";
        this->initialised = 1;
    };
    mEmojiService->loadCustomEmojis( success );
}

void RocketChatServerData::openPrivateChannelWith( const QString &pUsername )
{

    if ( mChannels->getChannelByName( pUsername ) ) {
        switchChannelByRid( pUsername );
    } else {
        mChannelService->openPrivateChannelWith( pUsername );
    }
}

//TODO: clear up memory when finished with unsetbinding
void RocketChatServerData::uploadFile( const QString &pChannelId, const QString &pFilePath )
{
    qDebug() << "uploading file: " << pFilePath;
    qDebug() << "channelId: " << pChannelId;

    if ( !pFilePath.isEmpty() && !pChannelId.isEmpty() ) {
        QUrl uri( pFilePath );
        QFile file( pFilePath );

        if ( file.open( QFile::ReadOnly ) ) {
            file.close();
        }

        FileUploader *uploader = new FileUploader( this, uri, pChannelId );
        std::function<void( void )> success = [ = ]() {
            emit fileUploadFinished( uploader->getFileId() );
            delete uploader;
        };
        connect( uploader, &FileUploader::ufsCreated, this, [ = ]( const QString & fileId ) {
            QJsonObject data;
            data[QStringLiteral( "fileId" )] = fileId;
            data[QStringLiteral( "filePath" )] = Utils::getPathPrefix() + pFilePath;
            data[QStringLiteral( "filename" )] = uri.fileName();
            mFileUploads[fileId] = uploader;
            QJsonDocument doc( data );
            emit fileuploadStarted( doc.toJson() );
        } );
        connect( uploader, &FileUploader::progressChanged, this, [ = ]( double progress ) {
            qDebug() << "progress changed";
            QJsonObject data;
            data[QStringLiteral( "fileId" )] = uploader->getFileId();
            data[QStringLiteral( "progress" )] = progress;
            QJsonDocument doc( data );
            emit fileUploadProgressChanged( doc.toJson() );
        } );
        connect( uploader, &FileUploader::error, this, [ = ]( const QString & errorString ) {
            emit error( errorString );
        } );

        try {
            uploader->upload( success );
        } catch ( std::exception e ) {
            qCritical() << e.what();
        }
    }

}

void RocketChatServerData::cancelUpload( const QString &pFileId )
{
    if ( mFileUploads.contains( pFileId ) ) {
        auto upload = mFileUploads[pFileId];
        upload->cancel();
        mFileUploads.remove( pFileId );
    }
}

void RocketChatServerData::getUserSuggestions( const QString &pTerm, const QString &pExeptions )
{
    QJsonArray params;
    QJsonObject firstParam;
    QJsonDocument jsonExceptions = QJsonDocument::fromJson( pExeptions.toUtf8() );

    QStringList users = pExeptions.split( ',' );
    QJsonArray exceptionEntries;

    for ( const auto &entry : users ) {
        exceptionEntries.append( entry );
    }

    //exceptionEntries.append( mUsername );
    firstParam[QStringLiteral( "exceptions" )] = exceptionEntries;
    firstParam[QStringLiteral( "term" )] = pTerm;
    params.append( firstParam );

    QSharedPointer<DDPSubscriptionRequest> subSuggestions( new DDPSubscriptionRequest( QStringLiteral( "userAutocomplete" ), params ) );
    std::function<void ( QJsonObject, MeteorDDP * )> success = [ = ]( QJsonObject pResponse, MeteorDDP * pDdp ) {

        if ( pDdp ) {
            if ( pResponse.contains( QStringLiteral( "msg" ) ) ) {
                if ( pResponse[QStringLiteral( "msg" )] == QStringLiteral( "ready" ) ) {
                    emit suggestionsReady( mAutocompleteRecords );
                    mAutocompleteRecords.clear();
                    QSharedPointer<DDPUnsubscriptionRequest> unSubSuggestions( new DDPUnsubscriptionRequest( subSuggestions->getFrame() ) );
                    mDdpApi->sendRequest( unSubSuggestions );
                    pDdp->unsetResponseBinding( subSuggestions->getFrame() );

                }
            }
        }
    };
    subSuggestions->setSuccess( success );
    mDdpApi->sendRequest( subSuggestions );
}

void RocketChatServerData::getRoomInformation( const QString &pRoomName )
{
    QJsonArray params;
    params.append( pRoomName );

    QSharedPointer<RocketChatSubscribeChannelRequest> roomSubscription( new RocketChatSubscribeChannelRequest( pRoomName ) );

    mDdpApi->sendRequest( roomSubscription );
}

void RocketChatServerData::getLoginMethods()
{
    DdpCallback success = [ = ]( QJsonObject pResponse, MeteorDDP * pDdpApi ) {
        Q_UNUSED( pResponse )
        Q_UNUSED( pDdpApi );
        qDebug() << "loginMethods Recieved";
    };
    QSharedPointer<RocketChatSubscribeLoginMethods> loginMethodsSubscription( new RocketChatSubscribeLoginMethods() );
    loginMethodsSubscription->setSuccess( success );
    mDdpApi->sendRequest( loginMethodsSubscription );
}

void RocketChatServerData::getServerInfo()
{
    RestRequestCallback success = [ = ]( QNetworkReply * pReply, QJsonObject data, RestApi * pApi ) {
        Q_UNUSED( pApi );
        Q_UNUSED( pReply );
        qDebug() << data;

        if ( data.contains( QStringLiteral( "info" ) ) ) {
            QJsonObject info = data[QStringLiteral( "info" )].toObject();

            if ( info.contains( QStringLiteral( "version" ) ) ) {
                QString version = info[QStringLiteral( "version" )].toString();
                QStringList verParts = version.split( '.' );
                std::get<0>( RocketChatServerConfig::serverVersion ) = verParts[0].toInt();
                std::get<1>( RocketChatServerConfig::serverVersion ) = verParts[1].toInt();
                std::get<2>( RocketChatServerConfig::serverVersion ) = verParts[2].toInt();
            }
        }
    };
    QSharedPointer<GetServerInfoRequest> serverInfoRequest( new GetServerInfoRequest( success, nullptr ) );
    mRestApi->sendRequest( serverInfoRequest );
}

void RocketChatServerData::leaveChannel( const QString &pId )
{
    mChannelService->leaveChannel( pId );
}

void RocketChatServerData::hideChannel( const QString &pId )
{
    mChannelService->hideRoom( pId );
}

void RocketChatServerData::reportAbusiveContent( const QString &pMessageId, const QString &pAuthor )
{
    mChannelService->reportContent( pMessageId, pAuthor );
}

QList<QSharedPointer<RocketChatUser> > RocketChatServerData::getUserOfChannel( const QString &pChannelId )
{
    QList<QSharedPointer<RocketChatUser>> users;
    users.reserve( 100 );

    if ( mChannels->contains( pChannelId ) && !mChannels->get( pChannelId ).isNull() ) {
        QSharedPointer<RocketChatChannel> channel = mChannels->get( pChannelId );
        users =  channel->getUsers().values();
    }

    return users;
}

RestApi *RocketChatServerData::getRestApi() const
{
    return mRestApi;
}

void RocketChatServerData::resume( void )
{
    mResumeOperation = true;
    mDdpApi->resume();
}

void RocketChatServerData::logout()
{
    qDebug() << "logged out";
    RestApiRequest logoutRequest = QSharedPointer<RestRequest>( new RestLogoutRequest );
    mResumeOperation = false;
    emit resetLoginMethods();
    mDdpApi->resume();
    // mRestApi->logout();
    mRestApi->sendRequest( logoutRequest );
    mChannels->clear();
    mUsers.clear();
    mLoginMethodRepo.clear();
    mStorage->wipeDb();
    this->mUsername = "";
    mDdpApi->setToken( "" );
    //    mRestApi->deleteLater();
    //    setRestApi( new RestApi( "https://" + mBaseUrl, mApiUri ) );
    //    mRestApi->moveToThread(QThread::currentThread());
    mLoggedIn = false;
    emit loggedOut( this->mServerId );
}

void RocketChatServerData::createChannel( const QString &pChannelName, const QStringList &pUsers, bool pReadonly )
{
    DdpCallback success = [ = ]( QJsonObject pResponse, MeteorDDP * pDdpApi ) {
        Q_UNUSED( pResponse );
        Q_UNUSED( pDdpApi );
        qDebug() << "created channel";
    };
    QSharedPointer<RocketChatCreateChannelRequest> request( new RocketChatCreateChannelRequest( pChannelName, pUsers, pReadonly ) );
    mDdpApi->sendRequest( request );
}

void RocketChatServerData::createPrivateGroup( const QString &pChannelName, const QStringList &pUsers, bool pReadonly )
{
    DdpCallback success = [ = ]( QJsonObject response, MeteorDDP * ddpApi ) {
        Q_UNUSED( response );
        Q_UNUSED( ddpApi );
        qDebug() << "created group";
    };
    QSharedPointer<RocketChatCreatePrivateGroupRequest> request( new RocketChatCreatePrivateGroupRequest( pChannelName, pUsers, pReadonly ) );
    mDdpApi->sendRequest( request );
}

void RocketChatServerData::addUsersToChannel( const QString &pChannelName, const QStringList &pUsers )
{
    for ( const auto &user : pUsers ) {
        DdpCallback success = [ = ]( QJsonObject response, MeteorDDP * ddpApi ) {
            Q_UNUSED( response );
            Q_UNUSED( ddpApi );
            qDebug() << "added users to channel";
        };

        if ( Q_LIKELY( !user.isEmpty() ) ) {
            QStringList usernames( user );
            QSharedPointer<RocketChatAddUsersToChannel > request( new RocketChatAddUsersToChannel( pChannelName, usernames ) );
            mDdpApi->sendRequest( request );
        }
    }
}

ChannelRepository *RocketChatServerData::getChannels() const
{
    return mChannels;
}
